## Written by Matias Bordese <matiasb@except.com.ar>
## (c) 2007 ifPeople http://ifpeople.net/

# This file is part of ifSearchSpellSuggest.
#
#     ifSearchSpellSuggest is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
#
#     ifSearchSpellSuggest is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with ifSearchSpellSuggest; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from AccessControl import ClassSecurityInfo
from Products.CMFCore.CMFCorePermissions import View, ModifyPortalContent

from Products.CMFCore.utils import getToolByName, UniqueObject
from OFS.SimpleItem import SimpleItem
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from Globals import InitializeClass

from string import join
import re

from config import TOOLNAME

class ifSearchSpellSuggestTool(UniqueObject, SimpleItem):
    """A tool to handle the search suggestions"""

    id = 'ifSearchSpellSuggest_tool'
    meta_type= TOOLNAME
    plone_tool = 1

    security = ClassSecurityInfo()

    def __init__(self):
        self._synForward = PersistentMapping({})
        self._synBackward = PersistentMapping({})

    security.declareProtected(View, 'getSynonymKeys')
    def getSynonymKeys(self):
        """Return a list of words whose synonym is known"""
        ret = self._synForward.keys()
        ret.sort()
        return ret

    security.declareProtected(View, 'getWordSynonym')
    def getWordSynonym(self, word):
        """Return a word synonym or None"""
        ret = None
        word = word.lower()
        if word in self._synForward.keys():
            ret = self._synForward[word]
        if word in self._synBackward.keys():
            ret = self._synBackward[word]
        return ret

    security.declareProtected(View, 'addWordSynonym')
    def addWordSynonym(self, word, synonym):
        """Add a word and its synonym"""
        word = word.lower()
        synonym = synonym.lower()
        self._synForward[word] = synonym
        self._synBackward[synonym] = word

    security.declareProtected(View, 'delSynonymFromWeb')
    def delSynonymFromWeb(self, word, RESPONSE=None):
        """Deletes a word and its synonym"""
        self.delWordSynonym(word)
        if RESPONSE:
            RESPONSE.redirect(self.absolute_url() +
                              '/ifSearchSpellSuggest_dict?portal_status_message=' + word + ' entry deleted')

    security.declareProtected(View, 'delWordSynonym')
    def delWordSynonym(self, word):
        """Deletes a word and its synonym"""
        if word in self._synForward.keys():
            synonym = self._synForward[word]
            del self._synForward[word]
            del self._synBackward[synonym]
        elif word in self._synBackward.keys():
            synonym = self._synBackward[word]
            del self._synForward[synonym]
            del self._synBackward[word]

    security.declareProtected(View, 'importWordSynonyms')
    def importWordSynonyms(self, import_file, RESPONSE=None):
        """Import synonyms from a plain text file with a pair word, synonym by line"""
        res = 'successfully'
        fileContent = import_file.read()
        lines = fileContent.splitlines()
        for l in lines:
            try:
                word, syn = re.findall('\w+', l)
                self.addWordSynonym(word, syn)
            except:
                res = 'with errors'
        if RESPONSE:
            RESPONSE.redirect(self.absolute_url() +
                              '/ifSearchSpellSuggest_dict?portal_status_message=file imported ' + res)

    security.declareProtected(View, 'exportWordSynonyms')
    def exportWordSynonyms(self, RESPONSE=None):
        """Export synonyms to a plain text file with a pair word, synonym by line"""
        if RESPONSE:
            ret = ''
            RESPONSE.setHeader('content-type', 'text/plain')
            words = self.getSynonymKeys()
            for w in words:
                syn = self.getWordSynonym(w)
                res = '%s, %s\n' % (w, syn)
                ret = ret + res
            return ret

    security.declareProtected(View, 'getSynonymSuggestion')
    def getSynonymSuggestion(self, searchterm):
        """Return a search suggestion based on the synonyms stored"""
        ret = ''
        terms = searchterm.lower().split()
        for t in terms:
            term = re.findall('\w+', t)
            if term: term = term[0]
            else: term = ''
            s = self.getWordSynonym(term)
            if term not in ['and', 'not', 'or'] and s:
                rep = re.sub('\w+', s, t)
                ret = ret + rep + ' '
            else:
                ret = ret + t + ' '
        if ret:
            ret = ret[:-1]
        if ret == searchterm.lower():
            ret = None
        return ret

    security.declareProtected(View, 'getSpellingSuggestion')
    def getSpellingSuggestion(self, searchterm):
        """Return a new spelling suggestion for search, if any"""
        ret = ''
        terms = searchterm.lower().split()
        for t in terms:
            term = re.findall('\w+', t)
            if term: term = term[0]
            else: term = ''
            s = self.suggestWordSpelling(term) #get all suggestions
            sw = None
            for i in s: # take the most frequent, but not repeat
                if i not in terms:
                    sw = i
                    break
            if term not in ['and', 'not', 'or'] and term not in s and sw:
                rep = re.sub('\w+', sw, t)
                ret = ret + rep + ' '
            else:
                ret = ret + t + ' '
        if ret:
            ret = ret[:-1]
        if ret == searchterm.lower():
            ret = None
        return ret

    security.declarePrivate('suggestWordSpelling')
    def suggestWordSpelling(self, pattern):
        """Return a spelling suggestion for the given word, based in the SearchableText indexed words;
           it returns a 0.8 near levenshtein distant word, the one with most occurences in the index;
           requires ST to be a textindexng"""
        ret = None
        try:
            ct = self.portal_catalog
            index = ct.Indexes['SearchableText'].index
            lexicon = index.getLexicon()
            pattern = unicode(pattern)
            words = lexicon.getSimiliarWords(pattern, 0.8)
            words.sort(key=lambda x: self.suggestWordSortKey(x[0]), reverse=True)
            ret = [w[0] for w in words]
        except:
            pass
        return ret

    security.declarePrivate('suggestWordSortKey')
    def suggestWordSortKey(self, word):
        """Return the word frequency in the index"""
        ct = self.portal_catalog
        index = ct.Indexes['SearchableText'].index
        lexicon_getWordId = index.getLexicon().getWordId
        stg = index.getStorage('SearchableText')

        wid = lexicon_getWordId(word)
        docids = stg.getDocumentsForWordId(wid)
        TF = len(docids)   # term frequency

        return TF

InitializeClass(ifSearchSpellSuggestTool)
